use crossterm::{
	event::{poll, read, Event, KeyCode},
	terminal, ExecutableCommand, cursor,
};
use std::panic;
use std::{
	io::{stdout, Result, Write, Stdout},
	time::Duration,
};


#[derive(Default)]
struct Nuudel {
	quitting: bool,
	cursorx: u16,
	cursory: u16,
	buffer: String,
}

fn main() -> Result<()> {
	// Setup panic handling
	let default_panic = panic::take_hook();
	panic::set_hook(Box::new(move |info| {
		let _ = Nuudel::restore_terminal();
		default_panic(info);
	}));

	// Basic stuff
	let mut nuudel = Nuudel::default();
	Nuudel::init_terminal()?;
	nuudel.run(&mut stdout())?;
	Nuudel::restore_terminal()?;

	Ok(())
}

impl Nuudel {

	pub fn run(&mut self, stdout: &mut Stdout) -> Result<()> {
		while !self.quitting {
			self.handle_events()?;
			self.draw_terminal(stdout)?;
		}
		Ok(())
	}

	fn init_terminal() -> Result<()> {
		terminal::enable_raw_mode()?;
		stdout().execute(terminal::EnterAlternateScreen)?;
		Ok(())
	}

	fn restore_terminal() -> Result<()> {
		stdout().execute(cursor::Show)?;
		stdout().execute(terminal::LeaveAlternateScreen)?;
		terminal::disable_raw_mode()?;
		Ok(())
	}
	fn quit(&mut self) {
		self.quitting = true;
	}

	fn handle_events(&mut self) -> Result<()> {
		if poll(Duration::from_secs(0))? {
			match read()? {
				Event::Key(event) => match event.code {
					KeyCode::Char('q') => {
						self.quit();
					},
					KeyCode::Char('a') => {
						print!("█");
					},
					KeyCode::Left => {
						if self.cursorx > 0 {
							self.cursorx -= 1;
							self.cursorx = self.cursorx.clamp(0, terminal::size()?.0-1);
						}
					},
					KeyCode::Right => {
						self.cursorx += 1;
						self.cursorx = self.cursorx.clamp(0, terminal::size()?.0-1);
					},
					KeyCode::Up => {
						if self.cursory > 0 {
							self.cursory -= 1;
						}
						self.cursory = self.cursory.clamp(0, terminal::size()?.1-1);
					},
					KeyCode::Down => {
						self.cursory += 1;
						self.cursory = self.cursory.clamp(0, terminal::size()?.1-1);
					},
					_ => {}
				},
				_ => {}
			}
		}
		Ok(())
	}

	fn draw_terminal(&mut self, stdout: &mut Stdout) -> Result<()> {
		//stdout.execute(terminal::BeginSynchronizedUpdate)?;
		stdout.execute(cursor::Hide)?;
		stdout.execute(cursor::MoveTo(0, 0))?;

		let mut y = 0;
		self.buffer = "\n".to_string();
		let buffer_lines = self.buffer.lines().count() as u16;

		while y < terminal::size()?.1 - buffer_lines {
			self.buffer.push_str("~\n");
			y += 1;
		}

		// Draw buffer one by one
		y = 0;
		let mut lines = self.buffer.lines();
		while y < terminal::size()?.1-1 {
			stdout.execute(terminal::Clear(terminal::ClearType::CurrentLine))?;
			let line = lines.next();
			if line.is_none() {
				break;
			}
			write!(stdout, "{}", line.unwrap())?;
			stdout.execute(cursor::MoveToNextLine(1))?;
			y += 1;
		}
		print!("Press Q to quit");

		//stdout.execute(terminal::EndSynchronizedUpdate)?;
		stdout.execute(cursor::MoveTo(self.cursorx, self.cursory))?;
		stdout.execute(cursor::Show)?;

		std::thread::sleep(Duration::from_millis(16));

		Ok(())
	}
}
